@@ -16,11 +16,20 @@
 #include <pthread.h>
 
 #define BUF_SIZE 2048
+#define MICROSECONDS_IN_SECOND 1000000
 
 #define FAKE_FRAME_GC    INT2FIX(0)
 #define FAKE_FRAME_MARK  INT2FIX(1)
 #define FAKE_FRAME_SWEEP INT2FIX(2)
 
+/*
+ * As of Ruby 3.0, it should be safe to read stack frames at any time
+ * See https://github.com/ruby/ruby/commit/0e276dc458f94d9d79a0f7c7669bde84abe80f21
+ */
+#if RUBY_API_VERSION_MAJOR < 3
+  #define USE_POSTPONED_JOB
+#endif
+
 static const char *fake_frame_cstrs[] = {
 	"(garbage collection)",
 	"(marking)",
@@ -46,6 +55,7 @@ static struct {
     VALUE interval;
     VALUE out;
     VALUE metadata;
+    int ignore_gc;
 
     VALUE *raw_samples;
     size_t raw_samples_len;
@@ -73,8 +83,8 @@ static struct {
 
 static VALUE sym_object, sym_wall, sym_cpu, sym_custom, sym_name, sym_file, sym_line;
 static VALUE sym_samples, sym_total_samples, sym_missed_samples, sym_edges, sym_lines;
-static VALUE sym_version, sym_mode, sym_interval, sym_raw, sym_metadata, sym_frames, sym_out, sym_aggregate, sym_raw_timestamp_deltas;
-static VALUE sym_state, sym_marking, sym_sweeping;
+static VALUE sym_version, sym_mode, sym_interval, sym_raw, sym_metadata, sym_frames, sym_ignore_gc, sym_out;
+static VALUE sym_aggregate, sym_raw_timestamp_deltas, sym_state, sym_marking, sym_sweeping;
 static VALUE sym_gc_samples, objtracer;
 static VALUE gc_hook;
 static VALUE rb_mStackProf;
@@ -88,6 +98,7 @@ stackprof_start(int argc, VALUE *argv, VALUE self)
     struct sigaction sa;
     struct itimerval timer;
     VALUE opts = Qnil, mode = Qnil, interval = Qnil, metadata = rb_hash_new(), out = Qfalse;
+    int ignore_gc = 0;
     int raw = 0, aggregate = 1;
 
     if (_stackprof.running)
@@ -99,6 +110,9 @@ stackprof_start(int argc, VALUE *argv, VALUE self)
 	mode = rb_hash_aref(opts, sym_mode);
 	interval = rb_hash_aref(opts, sym_interval);
 	out = rb_hash_aref(opts, sym_out);
+	if (RTEST(rb_hash_aref(opts, sym_ignore_gc))) {
+	    ignore_gc = 1;
+	}
 
 	VALUE metadata_val = rb_hash_aref(opts, sym_metadata);
 	if (RTEST(metadata_val)) {
@@ -115,6 +129,10 @@ stackprof_start(int argc, VALUE *argv, VALUE self)
     }
     if (!RTEST(mode)) mode = sym_wall;
 
+    if (!NIL_P(interval) && (NUM2INT(interval) < 1 || NUM2INT(interval) >= MICROSECONDS_IN_SECOND)) {
+        rb_raise(rb_eArgError, "interval is a number of microseconds between 1 and 1 million");
+    }
+
     if (!_stackprof.frames) {
 	_stackprof.frames = st_init_numtable();
 	_stackprof.overall_signals = 0;
@@ -151,6 +169,7 @@ stackprof_start(int argc, VALUE *argv, VALUE self)
     _stackprof.aggregate = aggregate;
     _stackprof.mode = mode;
     _stackprof.interval = interval;
+    _stackprof.ignore_gc = ignore_gc;
     _stackprof.metadata = metadata;
     _stackprof.out = out;
 
@@ -585,32 +604,31 @@ stackprof_record_gc_samples()
 static void
 stackprof_gc_job_handler(void *data)
 {
-    static int in_signal_handler = 0;
-    if (in_signal_handler) return;
     if (!_stackprof.running) return;
 
-    in_signal_handler++;
     stackprof_record_gc_samples();
-    in_signal_handler--;
 }
 
 static void
 stackprof_job_handler(void *data)
 {
-    static int in_signal_handler = 0;
-    if (in_signal_handler) return;
     if (!_stackprof.running) return;
 
-    in_signal_handler++;
     stackprof_record_sample();
-    in_signal_handler--;
 }
 
 static void
 stackprof_signal_handler(int sig, siginfo_t *sinfo, void *ucontext)
 {
+    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
+
     _stackprof.overall_signals++;
-    if (rb_during_gc()) {
+
+    if (!_stackprof.running) return;
+    if (!ruby_native_thread_p()) return;
+    if (pthread_mutex_trylock(&lock)) return;
+
+    if (!_stackprof.ignore_gc && rb_during_gc()) {
 	VALUE mode = rb_gc_latest_gc_info(sym_state);
 	if (mode == sym_marking) {
 	    _stackprof.unrecorded_gc_marking_samples++;
@@ -620,8 +638,13 @@ stackprof_signal_handler(int sig, siginfo_t *sinfo, void *ucontext)
 	_stackprof.unrecorded_gc_samples++;
 	rb_postponed_job_register_one(0, stackprof_gc_job_handler, (void*)0);
     } else {
+#ifdef USE_POSTPONED_JOB
 	rb_postponed_job_register_one(0, stackprof_job_handler, (void*)0);
+#else
+	stackprof_job_handler(0);
+#endif
     }
+    pthread_mutex_unlock(&lock);
 }
 
 static void
@@ -722,6 +745,7 @@ Init_stackprof(void)
     S(raw_timestamp_deltas);
     S(out);
     S(metadata);
+    S(ignore_gc);
     S(frames);
     S(aggregate);
     S(state);
